unit serial;


{
************************************************************

serial.pas

Low-Level Serial Communication
supports 8250, 16450 (no FIFO) and 16550 (16 byte FIFO)
using polling or interrupt

(c) Heiko Purnhagen 1995-1999
e-mail: purnhage@tnt.uni-hannover.de
WWW:    http://www.fet.uni-hannover.de/~purnhage/

This program may be copied according to the GNU General Public Licence.
See file COPYING for details.

950330   born
951029   ...
980613   looping to detect UART availability (Win95)
         timeout using 8253 ctr0 (1.2MHz)
980614   check 8253 operation, getclk
980615   fix ser save & restore
980616   serrxovr
980622   serint... routines for interrupt-based operation
         partly based on async.asm by David Kessner
         (e-mail: david@kessner.denver.co.us)
980623   seoi
980728   clkinit required on some notebooks (toshiba t2200sx / compaq aero)
         serintok, [int kick-start for serintrx]
980811   sertestfifo
         serselectcheck, serintinit(,,check,,) require successfull loop-back
         moved SEOI to end of int-service, [read iir to reset tx int]
980812   continue to read rbr even if rxbuf overflow (for Win95/DOSemu)
990823   bp7/dpmi modifications

************************************************************
}


{ serial interface connector                  }
{                                             }
{ D25  D9  in/out  signal                     }
{                                             }
{   1   -     -    FG   Frame Ground          }
{   2   3     O    TD   Transmit Data         }
{   3   2     I    RD   Receive Data          }
{   4   7     O    RTS  Request to Send       }
{   5   8     I    CTS  Clear to Send         }
{   6   6     I    DSR  Data Set Ready        }
{   7   5     -    SG   Signal Ground         }
{   8   1     I    DCD  Data Carrier Detected }
{  20   4     O    DTR  Data Terminal Ready   }
{  22   9     I    RI   Ring Indicator        }


{ +3V..+12V = active   = space = 0 }
{ -3V..-12V = inactive = mark  = 1 }


{ PC <-> PC cable:  SG  <-> SG  }
{                   TD  --> RD  }
{                   RD  <-- TD  }
{                   RTS --> CTS }
{                   CTS <-- RTS }
{                   DTR --> DSR }
{                        \> DCD }
{                   DSR <-- DTR }
{                   DCD </      }

{ loopback:  TD  --> RD  }
{            RTS --> CTS }
{            DTR --> DSR }
{                 \> DCD }

{ handshaking for PC <-> PC connections: (RTS<->CTS) }
{   transmit data only if CTS is active              }
{   set RTS active if waiting to receive data        }

{ I/O-address & IRQ:  COM1  $3f8  IRQ4 }
{                     COM2  $2f8  IRQ3 }
{                     COM3  $3e8  IRQ4 }
{                     COM4  $2e8  IRQ3 }


interface

uses
  crt,dos;   { for delay, setintvec/getintvec }

const
  clkfreq = 1.193180e6;   { pit 8253 cntr 0 clock freq. }
  clk1ms = longint(1193);

  sermaxbaud = longint(115200);   { max baud rate (divisor = 1) }

  longtimeout = 1000;   { 1 sec }
  shorttimeout = 100;   { 0.1 sec }

  parityno    = 0;   { parity modes }
  parityodd   = 1;
  parityeven  = 3;
  paritymark  = 5;
  parityspace = 7;

  serrecerrovr  = $02;   { receiver overrun error }
  serrecerrpar  = $04;   { receiver parity error }
  serrecerrfrm  = $08;   { receiver framing error }
  serrecerrbrk  = $10;   { receiver break detected }
  serrecerrfif  = $80;   { error in receiver fifo }
  serrecerrmask = $9e;   { uart rec error mask }
  serrecerrbuf  = $01;   { serint... buffer overflow }

  serbufsize = 1023;   { serint... rx & tx buffer size (must be 2^n-1) }


var
  sererror : boolean;     { true = error }
  sertimeout : boolean;   { true = timeout }
  serrecerr : integer;    { reciever error status  }
                          { (cleared only by init) }
                          { $00 = OK               }
                          { $?? = serrecerr???     }

{$ifdef msdos}
  timer : word absolute $0040:$006c;
  { counter using 8253 ctr0 & irq0 }
  { 1193180 Hz / 65536 = 18.206482 Hz }
{$else}
  { seg0040 is not a const in "dpmi"-mode ... }
  { timer : word absolute seg0040:$006c; }
{$endif}


procedure clkinit;
{ you shouldn't need this, but ... }

function clkcheck : boolean;
{ OK -> true }

procedure clkreset;

function clkget : longint;

procedure clkdelay (ms : word);


function serbase (com : integer) : word;
{ 0 if com not installed }

function sertestfifo (iobase : word) : boolean;
{ true if 16550 }

procedure sertestbase (var iobase : word; var is16550 : boolean);
{ iobase = 0 if not available }

procedure sertest (com : integer; var iobase : word; var is16550 : boolean);
{ iobase <> 0 if com available }

procedure serselect (com : integer);
{ sererror = false if select ok }

procedure serselectcheck (com : integer);
{ sererror = false if select ok }

procedure serinit (baud : longint; bits, stop, parity : integer);
{ sererror = false if init ok }

procedure serkeep;
{ disable serrestore & automatic serrestore on exit }

procedure serrestore;

procedure sersetfifo (usefifo, resetfifo : boolean);
{ init -> fifo disabled }

procedure sersethandshake (usetxcts, userxrts : boolean);
{ init -> handshake disabled }

procedure sersettimeout (txtimeout, rxtimeout : boolean; time : word);
{ init -> timeout enabled }

procedure sersetstatus (dtr, rts : boolean);
{ init -> rts & dtr inactive }

procedure sergetstatus (var cts, dsr, ri, dcd : boolean);

function sertxrdy : boolean;
{ true if transmitter ready }

procedure sertx (data : byte);
{ sertimeout = true if timeout }

function serrxrdy (setrts : boolean) : boolean;
{ true if data received }
{ RTS is activated only if setrts true }

function serrxovr : boolean;
{ true if receiver overrun }

function serrx : byte;
{ sertimeout = true if timeout }
{ serrecerr <> 0 if error }

function serrecerrstr (recerr : integer) : string;


function serintinit (com : integer; com5base : word; com5irq : integer;
                     baud : longint; bits, stop, parity : integer;
                     check, fifoenable : boolean;
                     var fifodetect : boolean) : boolean;
{ com1=$3f8,irq4 com2=$2f8,irq3 com3=$3e8,irq4, com4=$2e8,irq3 }
{ com5=com5base,com5irq (only irq0..irq7 supported) }
{ check=true -> require successfull loopback }
{ fifoenable=true -> enable 16550 fifo if available }
{ 16550 fifo detected -> fifodetect=true }
{ OK -> true / interrupt service installed  }

procedure serintrestore;
{ must be called to de-install interrupt service }

procedure serintclrbuf;
{ clear rx and tx buffers }

function serintrxbuf : integer;
{ num bytes in receiver buffer }

function serintrx : byte;
{ sererror = true if no data available }
{ serrecerr <> 0 if rec error in buffer or overflow }

function serinttxbuf : integer;
{ num bytes in transmitter buffer }

procedure serinttx (data : byte);
{ sererror = true if buffer overflow }

{ serint... routines can be combined with: }
{ sersetstatus, sergetstatus,  serrecerrstr }


implementation

const
             { uart 8250 / 16450 / 16550 registers }
  rbr = 0;   { receiver buffer (read) }
  thr = 0;   { transmitter hold (write) }
  ier = 1;   { interrupt enable }
  iir = 2;   { interrupt identification (read) }
  fcr = 2;   { fifo control (write) }
  lcr = 3;   { line control }
  mcr = 4;   { modem control }
  lsr = 5;   { line status }
  msr = 6;   { modem status }
  scr = 7;   { scratch }
  dll = 0;   { divisor latch (lsb) }
  dlm = 1;   { divisor latch (msb) }

                { pic 8259 registers (1st interrupt controller irq0..7) }
  imr = $21;    { read/write int mask reg (1=masked, d0=irq0) }
  ocw2 = $20;   { write control word (i.e. eoi) }
  eoi = $20;    { non specific end of interrupt }
  seoi = $60;    { specific end of interrupt +0..+7 }

{$ifdef msdos}
const
  seg0040 = $0040;
{$endif}

type
  wordptr = ^word;

var
  clkbufl, clkbufh : word;
  clk2bufl, clk2bufh : word;

  base : word;
  baseinit : boolean;
  txcts, rxrts : boolean;   { handshake }
  txto, rxto : boolean;   { timeout }
  totime : word;

  oriier,oriiir,orilcr,orimcr,oridll,oridlm : byte;

  serintok : boolean;
  irqint : integer;
  oldirqmask : byte;
  oldexitproc : pointer;
  oldirqproc : pointer;
  rxbuf : array [0..serbufsize] of byte;
  txbuf : array [0..serbufsize] of byte;
  rxput,rxget,txput,txget : word;


{ timer (pit 8253 cntr 0 clk 1.19318 MHz) }

procedure clkinit;
{ you shouldn't need this, but ... }
begin
  port[$0043] := $34;   { cntr 0: mode 2 16bit }
  port[$0040] := $00;
  port[$0040] := $00;
end;

function clkcheck : boolean;
{ OK -> true }
var
  t1,t2 : word;
begin
  clkcheck := true;
  port[$0043] := $00;   { cntr 0: latch }
  t1 := port[$0040];
  t1 := not(t1+(port[$0040] shl 8));
  port[$0043] := $00;   { cntr 0: latch }
  t2 := port[$0040];
  t2 := not(t2+(port[$0040] shl 8));
  if t1 = t2 then
  begin
    port[$0043] := $00;   { cntr 0: latch }
    t2 := port[$0040];
    t2 := not(t2+(port[$0040] shl 8));
    if t1 = t2 then
      clkcheck := false;
  end;
end;

procedure clkreset;
begin
  clkbufl := 0;
  clkbufh := 0;
end;

function clkget : longint;
  var
    t : word;
begin
  port[$0043] := $00;   { cntr 0: latch }
  t := port[$0040];
  t := not(t+(port[$0040] shl 8));
  if t < clkbufl then
    inc(clkbufh);
  clkbufl := t;
  clkget := clkbufl+(longint(clkbufh) shl 16);
end;

procedure clk2reset;
begin
  clk2bufl := 0;
  clk2bufh := 0;
end;

function clk2get : longint;
  var
    t : word;
begin
  port[$0043] := $00;   { cntr 0: latch }
  t := port[$0040];
  t := not(t+(port[$0040] shl 8));
  if t < clk2bufl then
    inc(clk2bufh);
  clk2bufl := t;
  clk2get := clk2bufl+(longint(clk2bufh) shl 16);
end;

procedure clkdelay (ms : word);
var
  t : longint;
begin
  if clkcheck then
  begin
    clk2reset;
    t := clk2get+longint(ms)*clk1ms;
    repeat
    until clk2get > t;
  end
  else
    delay (ms);   { tp6 delay as fallback }
end;


{ UART 8250 / 16450 / 16550 }

function serbase (com : integer) : word;
{ 0 if com not installed }
begin
  serbase := 0;
  if (com < 1) or (com > 4) then
    exit;
  serbase := wordptr(ptr(seg0040,2*(com-1)))^;
end;

function sertestfifo (iobase : word) : boolean;
{ true if 16550 }
var
  b : byte;
begin
  sertestfifo := false;
  if iobase <> 0 then
  begin
    b := port[iobase+iir];
    port[iobase+fcr] := 1;
    sertestfifo := (port[iobase+iir] and $80 = $80);
    port[iobase+fcr] := (b and $80) shr 7;
  end;
end;

procedure sertestbase (var iobase : word; var is16550 : boolean);
{ iobase = 0 if not available }
const
  tdmax = 100;
var
  b : byte;
  ok : boolean;
  oldier,oldiir,oldlcr,oldmcr,olddll,olddlm : byte;
  t0,td : longint;
  i : integer;
begin
  clkinit;
  if not clkcheck then
  begin
    iobase := 0;
    exit;
  end;
  oldiir := port[iobase+iir];
  oldlcr := port[iobase+lcr];
  oldmcr := port[iobase+mcr];
  port[iobase+lcr] := oldlcr and $7f;   { dlab=0 }
  oldier := port[iobase+ier];
  port[iobase+lcr] := oldlcr or $80;   { dlab=1 }
  olddll := port[iobase+dll];
  olddlm := port[iobase+dlm];
  port[iobase+dll] := $01;   { 115200 }
  port[iobase+dlm] := $00;
  port[iobase+lcr] := $03;   { 8n1 dlab=0 }
  port[iobase+ier] := $00;
  port[iobase+mcr] := oldmcr or $10;   { loop }
  port[iobase+fcr] := $06;   { disable and reset fifo }
  clk2reset;
  t0 := clk2get;
  repeat
    td := (clk2get-t0) div clk1ms;
  until (port[iobase+lsr] and $40 = $40) or (td > tdmax);
  for i := 1 to 16 do
    b := port[iobase+rbr];   { receive previous stuff }
  port[iobase+thr] := $ff;   { send }
  repeat
    td := (clk2get-t0) div clk1ms;
  until (port[iobase+lsr] and $41 = $41) or (td > tdmax);
  for i := 1 to 16 do
    b := port[iobase+rbr];   { receive previous stuff }
  port[iobase+thr] := $a5;   { send }
  repeat
    td := (clk2get-t0) div clk1ms;
  until (port[iobase+lsr] and $40 = $40) or (td > tdmax);
  repeat
    repeat
      td := (clk2get-t0) div clk1ms;
    until (port[iobase+lsr] and $01 = $01) or (td > tdmax);
    b := port[iobase+rbr];   { receive }
  until (b = $a5) or (td > tdmax);
  port[iobase+thr] := $5a;   { send }
  repeat
    td := (clk2get-t0) div clk1ms;
  until (port[iobase+lsr] and $41 = $41) or (td > tdmax);
  b := port[iobase+rbr];   { receive }
  if td > tdmax then
    ok := false
  else
    ok := (b = $5a);
  port[iobase+lcr] := oldlcr or $80;   { dlab=1 }
  port[iobase+dll] := olddll;
  port[iobase+dlm] := olddlm;
  port[iobase+lcr] := oldlcr and $7f;   { dlab=0 }
  port[iobase+ier] := oldier;
  port[iobase+lcr] := oldlcr;
  port[iobase+mcr] := oldmcr;
  if not ok then
    iobase := 0
  else
  begin
    port[iobase+fcr] := 1;
    is16550 := (port[iobase+iir] and $80 = $80);
  end;
  port[iobase+fcr] := (oldiir and $80) shr 7;
end;

procedure sertest (com : integer; var iobase : word; var is16550 : boolean);
{ iobase <> 0 if com available }
begin
  iobase := 0;
  is16550 := false;
  iobase := serbase(com);
  if iobase = 0 then
    exit;
  sertestbase (iobase,is16550);
end;

procedure serselect (com : integer);
{ sererror = false if select ok }
begin
  serrestore;
  sererror := true;
  base := serbase(com);
  if base <> 0 then
    sererror := false;
end;

procedure serselectcheck (com : integer);
{ sererror = false if select ok }
var
  b : boolean;
begin
  serrestore;
  sererror := true;
  sertest (com, base, b);
  if base <> 0 then
    sererror := false;
end;

procedure serinit (baud : longint; bits, stop, parity : integer);
{ sererror = false if init ok }
var
  b : byte;
  l : longint;
  i : integer;
begin
  sererror := true;
  baseinit := false;
  if base = 0 then
    exit;
  if (bits < 5) or (bits > 8) then
    exit;
  if (stop < 1) or (stop > 2) then
    exit;
  if (parity < 0) or (parity > 7) then
    exit;
  b := (bits-5) + (stop-1)*4 + parity*8;
  if baud <= 0 then
    exit;
  l := sermaxbaud div baud;
  if (sermaxbaud mod baud) >= (baud div 2) then
    l := l + 1;
  if (l < 1) or (l > $ffff) then
    exit;
  oriiir := port[base+iir];
  orilcr := port[base+lcr];
  orimcr := port[base+mcr];
  port[base+lcr] := orilcr and $7f;   { dlab=0 }
  oriier := port[base+ier];
  port[base+lcr] := orilcr or $80;   { dlab=1 }
  oridll := port[base+dll];
  oridlm := port[base+dlm];
  port[base+lcr] := $80;   { dlab = 1 }
  port[base+dll] := lo(word(l));
  port[base+dlm] := hi(word(l));
  port[base+lcr] := b;
  port[base+ier] := 0;   { no interrupts }
  port[base+fcr] := 0;   { fifo disabled }
  port[base+mcr] := 0;   { rts & dtr inactive }
  for i := 1 to 16 do   { empty fifo }
    b := port[base+rbr];
  b := port[base+lsr];   { reset lsr flags }
  b := port[base+msr];   { reset msr flags }
  serrecerr := 0;
  txcts := false;
  rxrts := false;
  txto := true;
  rxto := true;
  totime := longtimeout;
  baseinit := true;
  sererror := false;
end;

procedure serkeep;
{ disable serrestore & automatic serrestore on exit }
begin
  baseinit := false;
end;

{$f+}
procedure serrestore;
var
  i : integer;
  b : byte;
begin
  if not baseinit then
    exit;
  port[base+lcr] := orilcr or $80;   { dlab=1 }
  port[base+dll] := oridll;
  port[base+dlm] := oridlm;
  port[base+lcr] := orilcr and $7f;   { dlab=0 }
  port[base+ier] := oriier;
  port[base+lcr] := orilcr;
  port[base+mcr] := orimcr;
  port[base+fcr] := (oriiir and $80) shr 7;
  for i := 1 to 16 do   { empty fifo }
    b := port[base+rbr];
  baseinit := false;
end;
{$f-}

procedure sersetfifo (usefifo, resetfifo : boolean);
var
  b : byte;
begin
  if base = 0 then
    exit;
  b := 0;
  if usefifo then
  begin
    b := b + 1;
    if resetfifo then
      b := b + 6;
  end;
  port[base+fcr] := b;
end;

procedure sersethandshake (usetxcts, userxrts : boolean);
begin
  txcts := usetxcts;
  rxrts := userxrts;
end;

procedure sersettimeout (txtimeout, rxtimeout : boolean; time : word);
begin
  txto := txtimeout;
  rxto := rxtimeout;
  totime := time;
end;

procedure sersetstatus (dtr, rts : boolean);
var
  b : byte;
begin
  if base = 0 then
    exit;
  b := 0;
  if dtr then
    b := b + 1;
  if rts then
    b := b + 2;
  port[base+mcr] := b;
end;

procedure sergetstatus (var cts, dsr, ri, dcd : boolean);
var
  b : byte;
begin
  if base = 0 then
    exit;
  b := port[base+msr];
  cts := (b and $10) = $10;
  dsr := (b and $20) = $20;
  ri := (b and $40) = $40;
  dcd := (b and $80) = $80;
end;

function sertxrdy : boolean;
{ true if transmitter ready }
var
  b : byte;
begin
  sertxrdy := false;
  if base = 0 then
    exit;
  b := port[base+lsr];
  serrecerr := serrecerr or (b and serrecerrmask);
  if txcts then
    sertxrdy := ((b and $20) = $20) and ((port[base+msr] and $10) = $10)
  else
    sertxrdy := ((b and $20) = $20);
end;

procedure sertx (data : byte);
{ sertimeout = true if timeout }
var
  b : byte;
  t0,td : word;
begin
  if base = 0 then
    exit;
  clk2reset;
  t0 := clk2get;
  td := 0;
  repeat
    if txto then
      td := (clk2get-t0) div clk1ms;
    b := port[base+lsr];
    serrecerr := serrecerr or (b and serrecerrmask);
    if txcts and ((port[base+msr] and $10) = 0) then
      b := 0;
  until (td > totime) or ((b and $20) = $20);
  sertimeout := (td > totime);
  if not sertimeout then
    port[base+thr] := data;
end;

function serrxrdy (setrts : boolean) : boolean;
{ true if data received }
{ RTS is activated only if setrts true }
var
  b : byte;
  rdy : boolean;
begin
  serrxrdy := false;
  if base = 0 then
    exit;
  b := port[base+lsr];
  serrecerr := serrecerr or (b and serrecerrmask);
  rdy := ((b and $01) = $01);
  if rxrts and setrts and (not rdy) then
  begin
    inline ($fa);   { cli (disable interrupts) }
    port[base+mcr] := port[base+mcr] or $02;   { RTS active }
    b := port[base+lsr];
    serrecerr := serrecerr or (b and serrecerrmask);
    rdy := ((b and $01) = $01);
    port[base+mcr] := port[base+mcr] and $fd;   { RTS inactive }
    inline ($fb);   { sti (enable interrupts) }
  end;
  serrxrdy := rdy;
end;

function serrxovr : boolean;
{ true if receiver overrun }
var
  i : integer;
  b : byte;
begin
  serrxovr := false;
  if base = 0 then
    exit;
  if (serrecerr and serrecerrovr) <> 0 then
  begin
    serrxovr := true;
    b := port[base+rbr];
    if port[base+iir] and $80 = $80 then
      for i := 1 to 15 do
        b := port[base+rbr];
    serrecerr := serrecerr and not serrecerrovr;
  end;
end;

function serrx : byte;
{ sertimeout = true if timeout }
{ serrecerr <> 0 if error }
var
  b : byte;
  t0,td : word;
begin
  serrx := 0;
  if base = 0 then
    exit;
  clk2reset;
  t0 := clk2get;
  td := 0;
  inline ($fa);   { cli (disable interrupts) }
  b := port[base+lsr];
  serrecerr := serrecerr or (b and serrecerrmask);
  if rxrts and ((b and $01) = $00) then
    port[base+mcr] := port[base+mcr] or $02;   { RTS active }
  repeat
    if rxto then
      td := (clk2get-t0) div clk1ms;
    b := port[base+lsr];
    serrecerr := serrecerr or (b and serrecerrmask);
  until (td > totime) or ((b and $01) = $01);
  if rxrts then
    port[base+mcr] := port[base+mcr] and $fd;   { RTS inactive }
  inline ($fb);   { sti (enable interrupts) }
  sertimeout := (td > totime);
  if not sertimeout then
    serrx := port[base+rbr];
end;

function serrecerrstr (recerr : integer) : string;
var
  s : string;
begin
  if recerr = 0 then
    s := 'no error'
  else
  begin
    s := '';
    if (recerr and serrecerrovr) > 0 then
      s := s+'overrun error, ';
    if (recerr and serrecerrpar) > 0 then
      s := s+'parity error, ';
    if (recerr and serrecerrfrm) > 0 then
      s := s+'framing error, ';
    if (recerr and serrecerrbrk) > 0 then
      s := s+'break detected, ';
    if (recerr and serrecerrfif) > 0 then
      s := s+'error in fifo, ';
    if (recerr and serrecerrbuf) > 0 then
      s := s+'serint buffer overflow, ';
    if length(s) > 1 then
      delete (s,length(s)-1,2);
  end;
  serrecerrstr := s;
end;


{ serint... routines for interrupt-based operation }

{$f+}
procedure serintservice; interrupt;
{ interrupt service }
var
  b,r : byte;
begin
{  port[ocw2] := eoi;   { non specific eoi }
  port[ocw2] := seoi+irqint-8;   { specific eoi }
  if base <> 0 then
  begin
{    b := port[base+iir];   { to reset tx int }
    b := port[base+lsr];
    serrecerr := serrecerr or (b and serrecerrmask);
    if (b and $20) = $20 then
      if txput <> txget then
      begin
        port[base+thr] := txbuf[txget];
        txget := (txget+1) and serbufsize;
        if txput = txget then
          port[base+ier] := 1;   { enable only rx int }
      end;
    while (b and $01) = $01 do
    begin
      r := port[base+rbr];
      if ((rxput+1) and serbufsize) = rxget then
        serrecerr := serrecerr or serrecerrbuf
      else
      begin
        rxbuf[rxput] := r;
        rxput := (rxput+1) and serbufsize;
      end;
      b := port[base+lsr];
      serrecerr := serrecerr or (b and serrecerrmask);
    end;
{  b := port[base+ier];
{  port[base+ier] := 0;
{  port[base+ier] := b;   { generate interrupt edge }
  end;
{  port[ocw2] := seoi+irqint-8;   { specific eoi }
end;
{$f-}

procedure serintenable;
var
  b : byte;
begin
  if base = 0 then
    exit;
  oldirqmask := port[imr] or not (1 shl (irqint-$08));
  getintvec(irqint,oldirqproc);
  setintvec(irqint,@serintservice);
  port[imr] := port[imr] and not (1 shl (irqint-$08));   { unmask irq }
  port[base+ier] := 1;   { enable only rx int }
{  b := port[base+iir];   { to reset tx int }
{  port[base+ier] := 3;   { enable rx & tx int }
  port[base+mcr] := port[base+mcr] or $08;   { out2=1 to enable int }
  serintok := true;
end;

{$f+}
procedure serintdisable;
begin
  if serintok then
  begin
    if base <> 0 then
    begin
      port[base+mcr] := port[base+mcr] and $f7;   { out2=0 to disable int }
      port[base+ier] := 0;   { disable int }
    end;
    port[imr] := port[imr] or (1 shl (irqint-$08));   { mask irq }
    setintvec(irqint,oldirqproc);
    port[imr] := port[imr] and oldirqmask;
    serintok := false;
  end;
  serrestore;
end;
{$f-}

{$f+}
procedure serintexit;
{ de-install interrupt service on exit }
begin
  exitproc := oldexitproc;
  serintdisable;
end;
{$f-}

function serintinit (com : integer; com5base : word; com5irq : integer;
                     baud : longint; bits, stop, parity : integer;
                     check, fifoenable : boolean;
                     var fifodetect : boolean) : boolean;
{ com1=$3f8,irq4 com2=$2f8,irq3 com3=$3e8,irq4, com4=$2e8,irq3 }
{ com5=com5base,com5irq (only irq0..irq7 supported) }
{ fifoenable=true -> enable 16550 fifo if available }
{ 16550 fifo detected -> fifodetect=true }
{ OK -> true / interrupt service installed  }
var
  regs : registers;
begin
  serintrestore;
  serintinit := false;
  fifodetect := false;
  case com of
    1 :  begin
           base := $3f8;
           irqint := 4;
         end;
    2 :  begin
           base := $2f8;
           irqint := 3;
         end;
    3 :  begin
           base := $3e8;
           irqint := 4;
         end;
    4 :  begin
           base := $2e8;
           irqint := 3;
         end;
    5 :  begin
           base := com5base;
           irqint := com5irq;
         end;
    else begin
           base := 0;
           exit;
         end;
  end;
  if (irqint < 0) or (irqint > 7) then
  begin
    base := 0;
    exit;
  end;
  irqint := irqint+$08;
  if check then
  begin
    sertestbase (base,fifodetect);
    if base = 0 then
      exit;
  end
  else
    fifodetect := sertestfifo(base);
  serinit (baud,bits,stop,parity);
  if sererror then
  begin
    base := 0;
    exit;
  end;
  serintclrbuf;
  serintenable;
  if fifodetect and fifoenable then
    port[base+fcr] := $87;   { enable & clear fifos, rx trigger @ 8 bytes }
  intr(irqint,regs);   { kick-start interrupt service }
  serintinit := true;
end;

procedure serintrestore;
{ must be called to de-install interrupt service }
begin
  serintdisable;
end;

procedure serintclrbuf;
{ clear rx and tx buffers }
begin
  rxput := 0;
  rxget := 0;
  txput := 0;
  txget := 0;
  serrecerr := 0;
end;

function serintrxbuf : integer;
{ num bytes in receiver buffer }
begin
  serintrxbuf := 0;
  if not serintok then
    exit;
  serintrxbuf := (rxput-rxget) and serbufsize;
end;

function serintrx : byte;
{ sererror = true if no data available }
{ serrecerr <> 0 if rec error in buffer or overflow }
begin
  sererror := true;
  serintrx := 0;
  if not serintok then
    exit;
{  intr(irqint,regs);   { kick-start interrupt service }
  if rxget = rxput then
    exit;
  sererror := false;
  serintrx := rxbuf[rxget];
  rxget := (rxget+1) and serbufsize;
end;

function serinttxbuf : integer;
{ num bytes in transmitter buffer }
begin
  serinttxbuf := 0;
  if not serintok then
    exit;
  serinttxbuf := (txput-txget) and serbufsize;
end;

procedure serinttx (data : byte);
{ sererror = true if buffer overflow }
var
  regs : registers;
begin
  sererror := true;
  if not serintok then
    exit;
  if base = 0 then
    exit;
  if ((txput+1) and serbufsize) = txget then
    exit;
  sererror := false;
  txbuf[txput] := data;
  txput := (txput+1) and serbufsize;
  port[base+ier] := 3;   { enable rx & tx int }
  intr(irqint,regs);   { kick-start interrupt service }
end;


{ init unit }

begin
  base := 0;
  baseinit := false;
  serintok := false;
  oldexitproc := exitproc;
  exitproc := @serintexit;
end.

